
#ifndef QT_NO_IM
#include "qmultiinputcontext.h"
#include <qinputcontextfactory.h>
#include <qstringlist.h>
#include <qaction.h>
#include <qsettings.h>
#include <qmenu.h>

#include <stdlib.h>

QMultiInputContext::QMultiInputContext()
: QInputContext(), current(-1)
{
	QStringList keys = QInputContextFactory::keys();
	for (int i = keys.size()-1; i >= 0; --i)
		if (keys.at(i).contains(QLatin1String("imsw")))
			keys.removeAt(i);

	QString def = QLatin1String(getenv("QT_IM_MODULE"));
	if (def.isEmpty()) {
		QSettings settings;
		def = settings.value(QLatin1String("/qt/DefaultInputMethod"), QLatin1String("xim")).toString();
		printf("QMultiInputContext 3\n");
	}
	current = keys.indexOf(def);
	if (current < 0)
		current = 0;

	menu = new QMenu(tr("Select IM"));
	separator = new QAction(this);
	separator->setSeparator(true);

	QActionGroup *group = new QActionGroup(this);
	for (int i = 0; i < keys.size(); ++i) {
		slaves.append(QInputContextFactory::create(keys.at(i), this));
		QAction *a = menu->addAction(slaves.at(i)->identifierName());
		a->setCheckable(true);
		group->addAction(a);
		if (i == current)
			a->setChecked(true);
	}
	connect(group, SIGNAL(triggered(QAction*)), this, SLOT(changeSlave(QAction*)));

}

QMultiInputContext::~QMultiInputContext()
{
}


QString QMultiInputContext::identifierName()
{
	return (slave()) ? slave()->identifierName() : QLatin1String("");
}

QString QMultiInputContext::language()
{
	return (slave()) ? slave()->language() : QLatin1String("");
}


#if defined(Q_WS_X11)
bool QMultiInputContext::x11FilterEvent(QWidget *keywidget, XEvent *event)
{
	return (slave()) ? slave()->x11FilterEvent(keywidget, event) : false;
}
#endif // Q_WS_X11


bool QMultiInputContext::filterEvent(const QEvent *event)
{
	return (slave()) ? slave()->filterEvent(event) : false;
}

void QMultiInputContext::reset()
{
	if (slave())
		slave()->reset();
}

void QMultiInputContext::update()
{
	if (slave())
		slave()->update();
}

void QMultiInputContext::mouseHandler(int x, QMouseEvent *event)
{
	if (slave())
		slave()->mouseHandler(x, event);
}

QFont QMultiInputContext::font() const
{
	return (slave()) ? slave()->font() : QInputContext::font();
}

void QMultiInputContext::setFocusWidget(QWidget *w)
{
//	printf("setFocusWidget ! %x \n",w);	
	QInputContext::setFocusWidget(w); //???
	if (slave())
		slave()->setFocusWidget(w);
}

QWidget *QMultiInputContext::focusWidget() const
{
	return QInputContext::focusWidget();
}

void QMultiInputContext::widgetDestroyed(QWidget *w)
{
	if (slave())
		slave()->widgetDestroyed(w);
}

bool QMultiInputContext::isComposing() const
{
	return (slave()) ? slave()->isComposing() : false;
}

QList<QAction *> QMultiInputContext::actions()
{
	QList<QAction *> a = slave()->actions();
	a.append(separator);
	a.append(menu->menuAction());
	return a;
}

void QMultiInputContext::changeSlave(QAction *a)
{
	for (int i = 0; i < slaves.size(); ++i) {
		if (slaves.at(i)->identifierName() == a->text()) {
			current = i;
			return;
		}
	}
}
#endif
