#include <ctype.h>

#include <qapplication.h>
#include <qstring.h>
#include <qevent.h>
#include <qtextcodec.h>

#include <QInputMethodEvent>

#include <hangul.h>
#include "qinputcontexthangul.h"

HanjaTable* QInputContextHangul::hanjaTable = NULL;

static inline QString ucsToQString(const ucschar *ucs);

static bool onTransition(HangulInputContext* /* hic */,
		ucschar /* c */, const ucschar *str, void* /* data */)
{
	QTextCodec *codec = QTextCodec::codecForLocale();
	QString s = ucsToQString(str);
	return codec->canEncode(s);
}

static inline QString ucsToQString(const ucschar *ucs)
{
	QString str;

	if (ucs != NULL) {
		while (*ucs != 0)
			str += QChar(*ucs++);
	}
	return str;
}

QInputContextHangul::QInputContextHangul(const char* keyboard) :
	m_candidateList(NULL),
	m_mode(MODE_DIRECT),
	m_rect(0, 0, 0, 0)
{
	m_hic = hangul_ic_new(keyboard);
	//??
	hangul_ic_connect_callback(m_hic, "transition", (void*)onTransition, NULL);
	vkb=NULL;
}

QInputContextHangul::~QInputContextHangul()
{
	printf("QInputContextHangul destruct!\n");
//	if(vkb){ delete vkb; vkb=NULL;}
	if (m_hic != NULL)
		hangul_ic_delete(m_hic);
	printf("QInputContextHangul destruct! end\n");
}

QString QInputContextHangul::identifierName()
{
	return "Hangul";
}

QString QInputContextHangul::language()
{
	return "Hangul";
}

//???
QWidget *QInputContextHangul::focusWidget() const
{	
	return QInputContext::focusWidget();
}

//This virtual function is called when a state in the focus widget has changed. 
void QInputContextHangul::update()
{ 
	printf("update()\n");
}

void QInputContextHangul::widgetDestroyed( QWidget */*w*/ )
{
//	if(vkb)	delete vkb;
 	printf("widgetDestroyed()\n");
}

void QInputContextHangul::setFocusWidget(QWidget *w)
{
	if(w==NULL)
	{ 
		printf("hangul:lostFocusWidget !\n");	
		// unFocused
		return;
		//reset();
		setModeInfo(MODE_NONE);
		if (m_candidateList != NULL) {
			delete m_candidateList;
			m_candidateList = NULL;
		}
		targetWidget=NULL;
		return;
	}
  	Qt::WindowFlags flags = 0;
	
	if( targetWidget==w ){
		printf("setFocusWidget targetWidget!!\n");
		//targetWidget->clearFocus();
	   	targetWidget=NULL;
   	   return;
	}
	
	printf("hangul:setFocusWidget ! ObjName=%s ,className=%s \n",qPrintable(w->objectName()),w->metaObject()->className());


	if(strcmp(w->metaObject()->className(),"QLineEdit")!=0)
		return;

	if(w->objectName()=="vkbedit")
	{
		currentWidget=w;
		return ;
	}	
	else
	{
		targetWidget=w;
		flags=w->window()->windowFlags();

		//flags|=Qt::X11BypassWindowManagerHint;
		//flags=Qt::Popup;
		//w->window()->setWindowFlags(flags);
		//w->window()->setWindowFlags(0
		//
		if(vkb==NULL)
		{	
			vkb=new VirtualKeyboard(this);
			//		vkb->m_ui->vkbedit->setFocusProxy(w);
			///	m_ui->lineEdit->setAttribute(Qt::WA_InputMethodEnabled,false);
			//VirtualKeyboard *vkb=new VirtualKeyboard(0);
		}
		QLineEdit *e=(QLineEdit*)targetWidget;
		vkb->setText(e->text());
		vkb->show();
	}
	return;
	//if (framelessWindowCheckBox->isChecked())
   	//w->show();
	//return; 
	//w->window()->move(-100,-100);
	//printf("hangul:setFocusWidget ! %x \n",w);
	//printf("hangul:setFocusWidget ! %x \n",w->winId());	
	setModeInfo(m_mode);
//	if(vkb->isVisible()==false)
//	if(w->objectName()!="noime") return;
}

bool QInputContextHangul::isComposing () const
{
	printf("isComposing()\n");
	// const = 0
	return true;
}

// before lostFocusWidget(); 
void QInputContextHangul::reset()
{
	printf("reset()\n");
	
	const ucschar *flushed = hangul_ic_flush(m_hic);

	//???
	// we do not send preedit update IMEvent
	// because commit() send IM_End event and it remove preedit string
	// QString preeditString = getPreeditString();
	// updatePreedit(preeditString);

	QString commitString = ucsToQString(flushed);
	if (!commitString.isEmpty()) {
		commit(commitString);
	} 

}

QString QInputContextHangul::getPreeditString()
{
	return ucsToQString(hangul_ic_get_preedit_string(m_hic));
}

QString QInputContextHangul::getCommitString()
{
	return ucsToQString(hangul_ic_get_commit_string(m_hic));
}
#include <QTextFormat>
// composing text ready
void QInputContextHangul::updatePreedit(const QString &str)
{
	
	printf("preeditString()\n");
	QList<QInputMethodEvent::Attribute> attrs;

	QTextFormat xfmt=standardFormat ( QInputContext::SelectionFormat ) ;
//	QTextFormat xfmt;
	xfmt.setBackground(QColor (90,130,240));
	xfmt.setForeground(QColor (255,255,255));
	QInputMethodEvent::Attribute attr1( QInputMethodEvent::TextFormat, 0,1, xfmt );

	attrs.append(attr1);
	//Attribute::Attribute ( AttributeType type, int start, int length, QVariant value )
	
	QInputMethodEvent e(str,attrs);
	
//	e.setCommitString(str,0,str.length()); //replacement from,length
	sendEvent(e);
	
	return ;
	e.setCommitString(str,0,str.length()); //replacement from,length
	sendEvent(e);
	//if (!isComposing()) {
		////sendIMEvent(QEvent::IMStart);
	//}
	///sendIMEvent(QEvent::IMCompose, str, str.length());
}

void QInputContextHangul::commit(const QString &str)
{
	printf("commitString()\n");
	QInputMethodEvent e;
	e.setCommitString(str);
	sendEvent(e);
//	if (!isComposing()) {
	///	sendIMEvent(QEvent::IMStart);
//	}
	/// sendIMEvent(QEvent::IMEnd, str);
}

bool QInputContextHangul::isTriggerKey(const QKeyEvent *e)
{
	return (e->key() == Qt::Key_Space && (e->modifiers() & Qt::ShiftModifier))  || (e->key() == Qt::Key_Hangul);
}


// only hagul mode?
bool QInputContextHangul::backspace()
{
	printf("back\n");
	//if( hangul_ic_is_empty(m_hic)!=

	bool ret = hangul_ic_backspace(m_hic);
	if (ret) {
		QString str = getPreeditString();
		printf("preeditstring=%s\n",qPrintable(str));
		updatePreedit(str);
		///if (!isComposing()) sendIMEvent(QEvent::IMStart);
		//  	sendIMEvent(QEvent::IMCompose, str, str.length());
		//	else			sendIMEvent(QEvent::IMEnd, QString::null);
	}
	return ret;
}


bool QInputContextHangul::putAsciiKey(int key)
{
	static int key_prev=0;

	if (key == 0x08)
	{	
		if( backspace() == false)
		{
			// send keyevent to targetWidget
			//QKeyEvent ke( QEvent::KeyPress, Qt::Key_Backspace ,Qt::NoModifier);
			//virtual void keyPressEvent ( QKeyEvent * event )	
			QLineEdit *le=(QLineEdit*)currentWidget;
			le->backspace();
			return true;
		}
	}
	
	if (key== VK_FINISH) {
		reset(); //flush
		QLineEdit *e=	(QLineEdit*)targetWidget;
		e->setText(vkb->getText());	
		//targetWidget=NULL;
	}

	if (key== VK_ENG) {
		reset(); // flush Hangul
		m_mode = MODE_DIRECT;
		return true;
	}

	if (key== VK_HANGULE) {
		m_mode = MODE_HANGUL;
		return true;
	}

	printf("key ascii=%c \n",key);

	if (m_mode == MODE_HANGUL and ('A'<=key and key<='z')) 
	{
		if(key_prev==key)
		{
			switch(key)
			{
				case 'h': key='y';backspace();  break;
				case 'j': key='u';backspace();  break;
				case 'k': key='i'; backspace() ;break;
				case 'n': key='b'; backspace() ; break;
				case 'o': key='O'; backspace() ; break;
				case 'p': key='P'; backspace() ; break;
			//	if( backspace() == false)
				default :; 
			}
		}
		bool ret = hangul_ic_process(m_hic, key);

		QString commitString = getCommitString();
		if (!commitString.isEmpty())
			commit(commitString);
			
		//show prediting 
		QString preeditString = getPreeditString();
		if (!preeditString.isEmpty())
			updatePreedit(preeditString);
		
		key_prev=key;
		return ret;
	}
	
	if(key>0x21 and key <0x7d)
	{
		//English Alphabet
		commit(QString(QChar(key)));
	}
	return true;
}

bool QInputContextHangul::filterEvent(const QEvent *event)
{
	if (event->type() == QEvent::KeyRelease)
		return false;

	const QKeyEvent *keyevent = static_cast<const QKeyEvent*>(event);

	if (keyevent->key() == Qt::Key_Backspace)
	{	
		return backspace();
	}

	if (isTriggerKey(keyevent)) {
		if (m_mode == MODE_DIRECT) {
			m_mode = MODE_HANGUL;
		} else {
			//reset();
			m_mode = MODE_DIRECT;
		}
		//setModeInfo(m_mode);

		return true; // **********
	}

//	if (isCandidateKey(keyevent)) 	return popupCandidateList();
	/* ???
	if (keyevent->state() & Qt::ControlButton ||
			keyevent->state() & Qt::AltButton ||
			keyevent->state() & Qt::MetaButton) {
		reset();
		return false;
	} */

	int ascii = keyevent->key(); ///***
	
	if (keyevent->modifiers() & Qt::ShiftModifier)
		ascii = toupper(ascii);
	else
		ascii = tolower(ascii); 

	printf("keyboard ascii=%c \n",ascii);

	if (m_mode == MODE_HANGUL) {
		bool ret = hangul_ic_process(m_hic, ascii);
		
		//if(isComposing())

		QString commitString = getCommitString();
		if (!commitString.isEmpty())
		{
			commit(commitString);
		}	
		//??
		QString preeditString = getPreeditString();
		if (!preeditString.isEmpty())
		{	
			updatePreedit(preeditString);
		}
		return ret;
	}
	
	//else 
	if(ascii>0x21 and ascii <0x7d)
	{
		//English Alphabet
		commit(QString(QChar(ascii)));
		return true;
	}

	return false;
	//return true;
}
